# Copyright cocotb contributors
# Licensed under the Revised BSD License, see LICENSE for details.
# SPDX-License-Identifier: BSD-3-Clause


class InternalError(BaseException):
    """An error internal to scheduler. If you see this, report a bug!"""
